@@ -7,8 +7,18 @@ class DatasetMeta(Meta):
         super().__init__()
         self.tensors = []
         self.groups = []
+        self.tensor_names = {}
         self.hidden_tensors = []
 
+    @property
+    def visible_tensors(self):
+        return list(
+            filter(
+                lambda t: self.tensor_names[t] not in self.hidden_tensors,
+                self.tensor_names.keys(),
+            )
+        )
+
     @property
     def nbytes(self):
         # TODO: can optimize this
@@ -18,20 +28,22 @@ class DatasetMeta(Meta):
         d = super().__getstate__()
         d["tensors"] = self.tensors
         d["groups"] = self.groups
+        d["tensor_names"] = self.tensor_names
         d["hidden_tensors"] = self.hidden_tensors
         return d
 
-    def add_tensor(self, name, hidden=False):
-        if name not in self.tensors:
-            self.tensors.append(name)
+    def add_tensor(self, name, key, hidden=False):
+        if key not in self.tensors:
+            self.tensor_names[name] = key
+            self.tensors.append(key)
             if hidden:
-                self.hidden_tensors.append(name)
+                self.hidden_tensors.append(key)
             self.is_dirty = True
 
     def _hide_tensor(self, name):
-        assert name in self.tensors
+        assert name in self.tensor_names
         if name not in self.hidden_tensors:
-            self.hidden_tensors.append(name)
+            self.hidden_tensors.append(self.tensor_names[name])
             self.is_dirty = True
 
     def add_group(self, name):
@@ -40,10 +52,16 @@ class DatasetMeta(Meta):
             self.is_dirty = True
 
     def delete_tensor(self, name):
-        self.tensors.remove(name)
+        key = self.tensor_names.pop(name)
+        self.tensors.remove(key)
         self.is_dirty = True
 
     def delete_group(self, name):
         self.groups = list(filter(lambda g: not g.startswith(name), self.groups))
         self.tensors = list(filter(lambda t: not t.startswith(name), self.tensors))
         self.is_dirty = True
+
+    def rename_tensor(self, name, new_name):
+        key = self.tensor_names.pop(name)
+        self.tensor_names[new_name] = key
+        self.is_dirty = True
