@@ -1,6 +1,9 @@
 from hub.core.meta.encode.byte_positions import BytePositionsEncoder
 from hub.core.storage.hub_memory_object import HubMemoryObject
-from hub.core.serialize import serialize_sequence_encoder, deserialize_sequence_encoder
+from hub.core.serialize import (
+    serialize_sequence_or_creds_encoder,
+    deserialize_sequence_or_creds_encoder,
+)
 
 
 class SequenceEncoder(BytePositionsEncoder, HubMemoryObject):
@@ -9,7 +12,7 @@ class SequenceEncoder(BytePositionsEncoder, HubMemoryObject):
         instance = cls()
         if not buffer:
             return instance
-        version, ids = deserialize_sequence_encoder(buffer)
+        version, ids = deserialize_sequence_or_creds_encoder(buffer, "seq")
         if ids.nbytes:
             instance._encoded = ids
         instance.version = version
@@ -17,4 +20,6 @@ class SequenceEncoder(BytePositionsEncoder, HubMemoryObject):
         return instance
 
     def tobytes(self) -> memoryview:
-        return memoryview(serialize_sequence_encoder(self.version, self._encoded))
+        return memoryview(
+            serialize_sequence_or_creds_encoder(self.version, self._encoded)
+        )
