@@ -338,7 +338,7 @@ def test_add_to_non_empty_dataset(local_ds, scheduler, do_commit):
                 ds_out[index].label.numpy(), 15 * i * np.ones((1,))
             )
 
-    diff = ds_out.diff(as_dict=True)
+    diff = ds_out.diff(as_dict=True)["tensor"]
     change = {
         "image": {
             "data_updated": set(),
@@ -352,11 +352,15 @@ def test_add_to_non_empty_dataset(local_ds, scheduler, do_commit):
         },
     }
     if do_commit:
+        change["image"]["cleared"] = False
+        change["label"]["cleared"] = False
         change["image"]["created"] = False
         change["label"]["created"] = False
         change["image"]["data_added"] = [10, 610]
         change["label"]["data_added"] = [10, 610]
     else:
+        change["image"]["cleared"] = False
+        change["label"]["cleared"] = False
         change["image"]["created"] = True
         change["label"]["created"] = True
         change["image"]["data_added"] = [0, 610]
@@ -613,10 +617,11 @@ def test_inplace_transform(local_ds_generator):
             target = 2 if i % 2 == 0 else 3
             check_target_array(ds, i, target)
 
-        diff = ds.diff(as_dict=True)
+        diff = ds.diff(as_dict=True)["tensor"]
         change = {
             "img": {
                 "created": False,
+                "cleared": False,
                 "data_added": [0, 20],
                 "data_updated": set(),
                 "data_transformed_in_place": True,
@@ -624,6 +629,7 @@ def test_inplace_transform(local_ds_generator):
             },
             "label": {
                 "created": False,
+                "cleared": False,
                 "data_added": [0, 20],
                 "data_updated": set(),
                 "data_transformed_in_place": True,
@@ -868,3 +874,17 @@ def test_chunk_compression_bug(local_ds):
 
     for index in range(length):
         np.testing.assert_array_equal(ds.xyz[index].numpy(), xyz)
+
+
+@hub.compute
+def sequence_transform(inp, out):
+    out.x.append([np.ones(inp)] * inp)
+
+
+def test_sequence_htype_with_transform(local_ds):
+    ds = local_ds
+    with ds:
+        ds.create_tensor("x", htype="sequence")
+        sequence_transform().eval(list(range(1, 11)), ds, TRANSFORM_TEST_NUM_WORKERS)
+    for i in range(10):
+        np.testing.assert_array_equal(ds.x[i].numpy(), np.ones((i + 1, i + 1)))
