@@ -3,6 +3,7 @@ import hub
 from typing import Callable, List, Optional
 from itertools import repeat
 from hub.core.compute.provider import ComputeProvider
+from hub.core.storage.memory import MemoryProvider
 from hub.util.bugout_reporter import hub_reporter
 from hub.util.compute import get_compute_provider
 from hub.util.dataset import try_flushing
@@ -16,6 +17,7 @@ from hub.util.transform import (
     get_lengths_generated,
     get_old_chunk_paths,
     get_pbar_description,
+    process_transform_result,
     sanitize_workers_scheduler,
     store_data_slice,
     store_data_slice_with_pbar,
@@ -30,11 +32,12 @@ from hub.util.version_control import auto_checkout, load_meta
 
 
 class ComputeFunction:
-    def __init__(self, func, args, kwargs):
+    def __init__(self, func, args, kwargs, name: Optional[str] = None):
         """Creates a ComputeFunction object that can be evaluated using .eval or used as a part of a Pipeline."""
         self.func = func
         self.args = args
         self.kwargs = kwargs
+        self.name = self.func.__name__ if name is None else name
 
     def eval(
         self,
@@ -44,6 +47,7 @@ class ComputeFunction:
         scheduler: str = "threaded",
         progressbar: bool = True,
         skip_ok: bool = False,
+        check_lengths: bool = True,
     ):
         """Evaluates the ComputeFunction on data_in to produce an output dataset ds_out.
 
@@ -59,7 +63,7 @@ class ComputeFunction:
             progressbar (bool): Displays a progress bar if True (default).
             skip_ok (bool): If True, skips the check for output tensors generated. This allows the user to skip certain tensors in the function definition.
                 This is especially useful for inplace transformations in which certain tensors are not modified. Defaults to False.
-
+            check_lengths (bool): If True, checks whether ds_out has tensors of same lengths initially.
 
         Raises:
             InvalidInputDataError: If data_in passed to transform is invalid. It should support \__getitem__ and \__len__ operations. Using scheduler other than "threaded" with hub dataset having base storage as memory as data_in will also raise this.
@@ -69,7 +73,9 @@ class ComputeFunction:
         """
 
         pipeline = Pipeline([self])
-        pipeline.eval(data_in, ds_out, num_workers, scheduler, progressbar, skip_ok)
+        pipeline.eval(
+            data_in, ds_out, num_workers, scheduler, progressbar, skip_ok, check_lengths
+        )
 
     def __call__(self, sample_in):
         return self.func(sample_in, *self.args, **self.kwargs)
@@ -91,6 +97,7 @@ class Pipeline:
         scheduler: str = "threaded",
         progressbar: bool = True,
         skip_ok: bool = False,
+        check_lengths: bool = True,
     ):
         """Evaluates the pipeline on data_in to produce an output dataset ds_out.
 
@@ -106,6 +113,7 @@ class Pipeline:
             progressbar (bool): Displays a progress bar if True (default).
             skip_ok (bool): If True, skips the check for output tensors generated. This allows the user to skip certain tensors in the function definition.
                 This is especially useful for inplace transformations in which certain tensors are not modified. Defaults to False.
+            check_lengths (bool): If True, checks whether ds_out has tensors of same lengths initially.
 
         Raises:
             InvalidInputDataError: If data_in passed to transform is invalid. It should support \__getitem__ and \__len__ operations. Using scheduler other than "threaded" with hub dataset having base storage as memory as data_in will also raise this.
@@ -130,7 +138,8 @@ class Pipeline:
             data_in = get_dataset_with_zero_size_cache(data_in)
 
         target_ds = data_in if overwrite else ds_out
-        check_transform_ds_out(target_ds, scheduler)
+
+        check_transform_ds_out(target_ds, scheduler, check_lengths)
 
         # if overwrite then we've already flushed and autocheckecked out data_in which is target_ds now
         if not overwrite:
@@ -144,6 +153,10 @@ class Pipeline:
         initial_autoflush = target_ds.storage.autoflush
         target_ds.storage.autoflush = False
         progress_end_args = {"compute_id": compute_id, "progress": 100, "end": True}
+
+        if not check_lengths:
+            skip_ok = True
+
         try:
             self.run(
                 data_in,
@@ -157,7 +170,7 @@ class Pipeline:
             target_ds._send_compute_progress(**progress_end_args, status="success")
         except Exception as e:
             target_ds._send_compute_progress(**progress_end_args, status="failed")
-            raise TransformError(e)
+            raise TransformError(e) from e
         finally:
             compute_provider.close()
             if overwrite:
@@ -182,34 +195,42 @@ class Pipeline:
         """
         slices = create_slices(data_in, num_workers)
         storage = get_base_storage(target_ds.storage)
-        tensors = list(target_ds.tensors)
-        tensors = [target_ds.tensors[t].key for t in tensors]
+        visible_tensors = list(target_ds.tensors)
+        visible_tensors = [target_ds[t].key for t in visible_tensors]
+
+        tensors = list(target_ds._tensors())
+        tensors = [target_ds[t].key for t in tensors]
         group_index = target_ds.group_index
         version_state = target_ds.version_state
-        args = (storage, group_index, tensors, self, version_state, skip_ok)
-        map_inp = zip(slices, repeat(args))
+        if isinstance(storage, MemoryProvider):
+            storages = [storage] * len(slices)
+        else:
+            storages = [storage.copy() for _ in slices]
+        args = (
+            group_index,
+            tensors,
+            visible_tensors,
+            self,
+            version_state,
+            target_ds.link_creds,
+            skip_ok,
+        )
+        map_inp = zip(slices, storages, repeat(args))
 
         if progressbar:
             desc = get_pbar_description(self.functions)
-            metas_and_encoders = compute.map_with_progressbar(
+            result = compute.map_with_progressbar(
                 store_data_slice_with_pbar,
                 map_inp,
                 total_length=len(data_in),
                 desc=desc,
             )
         else:
-            metas_and_encoders = compute.map(store_data_slice, map_inp)
-
-        (
-            all_tensor_metas,
-            all_chunk_id_encoders,
-            all_tile_encoders,
-            all_chunk_commit_sets,
-            all_commit_diffs,
-        ) = zip(*metas_and_encoders)
+            result = compute.map(store_data_slice, map_inp)
+        result = process_transform_result(result)
 
         all_num_samples, all_tensors_generated_length = get_lengths_generated(
-            all_tensor_metas, tensors
+            result["tensor_metas"], tensors
         )
 
         check_lengths(all_tensors_generated_length, skip_ok)
@@ -220,16 +241,7 @@ class Pipeline:
 
         old_chunk_paths = get_old_chunk_paths(target_ds, generated_tensors, overwrite)
         merge_all_meta_info(
-            target_ds,
-            storage,
-            generated_tensors,
-            overwrite,
-            all_commit_diffs,
-            all_tile_encoders,
-            all_num_samples,
-            all_tensor_metas,
-            all_chunk_id_encoders,
-            all_chunk_commit_sets,
+            target_ds, storage, generated_tensors, overwrite, all_num_samples, result
         )
         delete_overwritten_chunks(old_chunk_paths, storage, overwrite)
 
@@ -244,31 +256,41 @@ def compose(functions: List[ComputeFunction]):  # noqa: DAR101, DAR102, DAR201,
 
     The __eval__ method evaluates the pipeline/transform function.
 
-    It has the following arguments:-
+    It has the following arguments:
 
-    - data_in: Input passed to the transform to generate output dataset.
-    It should support \__getitem__ and \__len__. This can be a Hub dataset.
-    - ds_out (Dataset, optional): The dataset object to which the transform will get written.
-    If this is not provided, data_in will be overwritten if it is a Hub dataset, otherwise error will be raised.
-    It should have all keys being generated in output already present as tensors.
-    It's initial state should be either:-
-        - Empty i.e. all tensors have no samples. In this case all samples are added to the dataset.
-        - All tensors are populated and have sampe length. In this case new samples are appended to the dataset.
-    - num_workers (int): The number of workers to use for performing the transform.
-    Defaults to 0. When set to 0, it will always use serial processing, irrespective of the scheduler.
-    - scheduler (str): The scheduler to be used to compute the transformation.
-    Supported values include: 'serial', 'threaded', 'processed' and 'ray'. Defaults to 'threaded'.
-    - progressbar (bool): Displays a progress bar if True (default).
-    - skip_ok (bool): If True, skips the check for output tensors generated. This allows the user to skip certain tensors in the function definition.
-    This is especially useful for inplace transformations in which certain tensors are not modified. Defaults to False.
+    - `data_in`: Input passed to the transform to generate output dataset.
+        - It should support \__getitem__ and \__len__. This can be a Hub dataset.
 
-    It raises the following errors:-
+    - `ds_out (Dataset, optional)`: The dataset object to which the transform will get written.
+        - If this is not provided, data_in will be overwritten if it is a Hub dataset, otherwise error will be raised.
+        - It should have all keys being generated in output already present as tensors.
+        - It's initial state should be either:
+            - Empty i.e. all tensors have no samples. In this case all samples are added to the dataset.
+            - All tensors are populated and have sampe length. In this case new samples are appended to the dataset.
 
-    - InvalidInputDataError: If data_in passed to transform is invalid. It should support \__getitem__ and \__len__ operations. Using scheduler other than "threaded" with hub dataset having base storage as memory as data_in will also raise this.
-    - InvalidOutputDatasetError: If all the tensors of ds_out passed to transform don't have the same length. Using scheduler other than "threaded" with hub dataset having base storage as memory as ds_out will also raise this.
-    - TensorMismatchError: If one or more of the outputs generated during transform contain different tensors than the ones present in 'ds_out' provided to transform.
-    - UnsupportedSchedulerError: If the scheduler passed is not recognized. Supported values include: 'serial', 'threaded', 'processed' and 'ray'.
-    - TransformError: All other exceptions raised if there are problems while running the pipeline.
+    - `num_workers (int)`: The number of workers to use for performing the transform.
+        - Defaults to 0. When set to 0, it will always use serial processing, irrespective of the scheduler.
+
+    - `scheduler (str)`: The scheduler to be used to compute the transformation.
+        - Supported values include: 'serial', 'threaded', 'processed' and 'ray'. Defaults to 'threaded'.
+
+    - `progressbar (bool)`: Displays a progress bar if True (default).
+
+    - `skip_ok (bool)`: If True, skips the check for output tensors generated.
+        - This allows the user to skip certain tensors in the function definition.
+        - This is especially useful for inplace transformations in which certain tensors are not modified. Defaults to False.
+
+    It raises the following errors:
+
+    - `InvalidInputDataError`: If data_in passed to transform is invalid. It should support \__getitem__ and \__len__ operations. Using scheduler other than "threaded" with hub dataset having base storage as memory as data_in will also raise this.
+
+    - `InvalidOutputDatasetError`: If all the tensors of ds_out passed to transform don't have the same length. Using scheduler other than "threaded" with hub dataset having base storage as memory as ds_out will also raise this.
+
+    - `TensorMismatchError`: If one or more of the outputs generated during transform contain different tensors than the ones present in 'ds_out' provided to transform.
+
+    - `UnsupportedSchedulerError`: If the scheduler passed is not recognized. Supported values include: 'serial', 'threaded', 'processed' and 'ray'.
+
+    - `TransformError`: All other exceptions raised if there are problems while running the pipeline.
     """
     if not functions:
         raise HubComposeEmptyListError
@@ -280,11 +302,16 @@ def compose(functions: List[ComputeFunction]):  # noqa: DAR101, DAR102, DAR201,
 
 def compute(
     fn,
+    name: Optional[str] = None,
 ) -> Callable[..., ComputeFunction]:  # noqa: DAR101, DAR102, DAR201, DAR401
     """Compute is a decorator for functions.
-    The functions should have atleast 2 argument, the first two will correspond to sample_in and samples_out.
+
+    The functions should have atleast 2 argument, the first two will correspond to `sample_in` and `samples_out`.
+
     There can be as many other arguments as required.
+
     The output should be appended/extended to the second argument in a hub like syntax.
+
     Any value returned by the fn will be ignored.
 
     Example::
@@ -305,34 +332,44 @@ def compute(
 
     The __eval__ method evaluates the pipeline/transform function.
 
-    It has the following arguments:-
+    It has the following arguments:
 
-    - data_in: Input passed to the transform to generate output dataset.
-    It should support \__getitem__ and \__len__. This can be a Hub dataset.
-    - ds_out (Dataset, optional): The dataset object to which the transform will get written.
-    If this is not provided, data_in will be overwritten if it is a Hub dataset, otherwise error will be raised.
-    It should have all keys being generated in output already present as tensors.
-    It's initial state should be either:-
-        - Empty i.e. all tensors have no samples. In this case all samples are added to the dataset.
-        - All tensors are populated and have sampe length. In this case new samples are appended to the dataset.
-    - num_workers (int): The number of workers to use for performing the transform.
-    Defaults to 0. When set to 0, it will always use serial processing, irrespective of the scheduler.
-    - scheduler (str): The scheduler to be used to compute the transformation.
-    Supported values include: 'serial', 'threaded', 'processed' and 'ray'. Defaults to 'threaded'.
-    - progressbar (bool): Displays a progress bar if True (default).
-    - skip_ok (bool): If True, skips the check for output tensors generated. This allows the user to skip certain tensors in the function definition.
-    This is especially useful for inplace transformations in which certain tensors are not modified. Defaults to False.
+    - `data_in`: Input passed to the transform to generate output dataset.
+        - It should support \__getitem__ and \__len__. This can be a Hub dataset.
 
-    It raises the following errors:-
+    - `ds_out (Dataset, optional)`: The dataset object to which the transform will get written.
+        - If this is not provided, data_in will be overwritten if it is a Hub dataset, otherwise error will be raised.
+        - It should have all keys being generated in output already present as tensors.
+        - It's initial state should be either:
+            - Empty i.e. all tensors have no samples. In this case all samples are added to the dataset.
+            - All tensors are populated and have sampe length. In this case new samples are appended to the dataset.
 
-    - InvalidInputDataError: If data_in passed to transform is invalid. It should support \__getitem__ and \__len__ operations. Using scheduler other than "threaded" with hub dataset having base storage as memory as data_in will also raise this.
-    - InvalidOutputDatasetError: If all the tensors of ds_out passed to transform don't have the same length. Using scheduler other than "threaded" with hub dataset having base storage as memory as ds_out will also raise this.
-    - TensorMismatchError: If one or more of the outputs generated during transform contain different tensors than the ones present in 'ds_out' provided to transform.
-    - UnsupportedSchedulerError: If the scheduler passed is not recognized. Supported values include: 'serial', 'threaded', 'processed' and 'ray'.
-    - TransformError: All other exceptions raised if there are problems while running the pipeline.
+    - `num_workers (int)`: The number of workers to use for performing the transform.
+        - Defaults to 0. When set to 0, it will always use serial processing, irrespective of the scheduler.
+
+    - `scheduler (str)`: The scheduler to be used to compute the transformation.
+        - Supported values include: 'serial', 'threaded', 'processed' and 'ray'. Defaults to 'threaded'.
+
+    - `progressbar (bool)`: Displays a progress bar if True (default).
+
+    - `skip_ok (bool)`: If True, skips the check for output tensors generated.
+        - This allows the user to skip certain tensors in the function definition.
+        - This is especially useful for inplace transformations in which certain tensors are not modified. Defaults to False.
+
+    It raises the following errors:
+
+    - `InvalidInputDataError`: If data_in passed to transform is invalid. It should support \__getitem__ and \__len__ operations. Using scheduler other than "threaded" with hub dataset having base storage as memory as data_in will also raise this.
+
+    - `InvalidOutputDatasetError`: If all the tensors of ds_out passed to transform don't have the same length. Using scheduler other than "threaded" with hub dataset having base storage as memory as ds_out will also raise this.
+
+    - `TensorMismatchError`: If one or more of the outputs generated during transform contain different tensors than the ones present in 'ds_out' provided to transform.
+
+    - `UnsupportedSchedulerError`: If the scheduler passed is not recognized. Supported values include: 'serial', 'threaded', 'processed' and 'ray'.
+
+    - `TransformError`: All other exceptions raised if there are problems while running the pipeline.
     """
 
     def inner(*args, **kwargs):
-        return ComputeFunction(fn, args, kwargs)
+        return ComputeFunction(fn, args, kwargs, name)
 
     return inner
